from django.apps import AppConfig


class StoreConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'hostore'
    verbose = 'Holmium Load Curve Store'
    label = 'hostore'
