# Generated by Django 5.0.2 on 2024-02-21 11:40

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TestDataStore',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('client_id', models.IntegerField()),
                ('prm', models.CharField(max_length=30)),
                ('last_modified', models.DateTimeField(auto_now=True)),
                ('data', models.BinaryField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
                'indexes': [models.Index(fields=['prm', 'client_id'], name='store_testd_prm_91e86e_idx')],
                'unique_together': {('prm', 'client_id')},
            },
        ),
    ]
