# Generated by Django 5.0.2 on 2024-05-03 13:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0003_alter_testdatastore_unique_together_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='testdatastore',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, null=True),
        ),
    ]
