# Generated by Django 5.0.2 on 2024-04-03 20:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0002_rename_store_testd_prm_91e86e_idx_hostore_tes_prm_897468_idx'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='testdatastore',
            unique_together=set(),
        ),
        migrations.AddField(
            model_name='testdatastore',
            name='version',
            field=models.IntegerField(default=0),
        ),
        migrations.AlterUniqueTogether(
            name='testdatastore',
            unique_together={('prm', 'client_id', 'version')},
        ),
    ]
